from idict.parameter.ifunctionspace import iFunctionSpace

compose = iFunctionSpace.fromfunctions

# TODO (minor): apply(fin="f", input=["X", "y"], output=["z"])     // need detection of multidynamic input (and output?)
